// FrameLoopPaint class for LoopPaint lab
// This class creates the frame for the plot of the loops

// (c) J. Dean Brock, Rebecca F. Bruce, 2001, 2004
// Department of Computer Science
// University of North Carolina at Asheville

package edu.unca.cs.csci201.LabAids;

import java.awt.* ;

/**
 * Creates the frame and manages the plots for the
 * LoopPaint lab.
 */
public class FrameLoopPaint {
    
    edu.unca.cs.csci201.LabAids.PanelLoopPaint realPanel ;
    javax.swing.JFrame realFrame ;
    
    // Number of objects to plot
    private int numObjects ;
    
    int numPlotFunc = 0 ;
    private SpecLoopPaint activeArtist ;
    
    /**
     * Constructor for FrameLoopPaint
     * @param title Title for the frame
     * @param width Width of panel <B>within</B> frame
     * @param height Height of panel <B>within</B> frame
     */
    public FrameLoopPaint(String title, int width, int height) {
        realPanel = 
              new edu.unca.cs.csci201.LabAids.PanelLoopPaint(width, height) ;
        realFrame = 
              edu.unca.cs.LabAids.FrameMaker.Component2Frame(realPanel, title) ;
    }
    
    private void INTaddPainter(SpecLoopPaint plotObject, String plotName) {
        final SpecLoopPaint thPlotObject = plotObject ;
        final String thPlotName = plotName ;
        Runnable thAddPainter = new Runnable() {
            public void run() {
                realPanel.addPainter(thPlotObject, thPlotName) ;
                realFrame.pack() ;
            }
        } ;
        try {
            javax.swing.SwingUtilities.invokeAndWait(thAddPainter) ;
        } catch (Exception e) {
            e.printStackTrace() ;
        }
        // javax.swing.SwingUtilities.invokeLater(thAddPainter) ;
    }

     /**
     * Adds a plot function with a name
     * @param paintObj Class used to tile the plot
     * @param paintNam Name for this plot
     */
    public void addPainter(SpecLoopPaint paintObj, String paintNam) {
        ++numObjects ;
        INTaddPainter(paintObj, paintNam) ;
    }
    
    /**
     * Adds a plot function with reasonable name
     * @param paintObj Class used to tile the plot
     */
    public void addPainter(SpecLoopPaint paintObj) {
        ++numObjects ;
        INTaddPainter(paintObj, "Function " + numObjects) ;
    }
    
    private class PainterLoopPaint implements edu.unca.cs.LabAids.Painter {
        SpecLoopPaint painterLP ;
        
        private class SyncStateData {
            public boolean[] progOutputs ;
            public boolean runFailure = false ;
            public int badIndex = -1 ;
            public Exception runException = null ;
            public boolean runComplete = false ;
        } ;
        
        SyncStateData progState = new SyncStateData() ;
        
        private class PainterSwitcher implements Switcher {
            public void TurnOn(int i) {
                synchronized(progState) {
                    if (0<=i && i< progState.progOutputs.length && !progState.runFailure)
                        progState.progOutputs[i] = true ;
                    else {
                        progState.runFailure = true ;
                        progState.badIndex = i ;
                    }
                }
            }
        }
        
        public PainterLoopPaint(SpecLoopPaint NPLP) {
            painterLP = NPLP ;
            
            Runnable computeNums = new Runnable() {
                public void run() {
                    try {
                        int numberOutputs = painterLP.size() ;
                        if (numberOutputs<=0) {
                            synchronized(progState) {
                                progState.runFailure = true ;
                                progState.runComplete = true ;
                            }
                        } else {
                            synchronized(progState) {
                                progState.progOutputs = new boolean[numberOutputs] ;
                            }
                            painterLP.TestFunc(new PainterSwitcher()) ;
                            synchronized(progState) {
                                progState.runComplete = true ;
                            }
                        }
                    } catch (Exception E) {
                        synchronized(progState) {
                            progState.runFailure = true ;
                            progState.runException = E ;
                            progState.runComplete = true ;
                        }
                    }
                }         
            } ;
            Thread computeThreads = new Thread(computeNums) ;
            // Let him run and hope he's done in time
            computeThreads.start() ;
        }
        
        public void paintCanvas(java.awt.Graphics g, javax.swing.JComponent panel) {
            g = g.create() ;
            // super.paint( g ) ;
            
            synchronized(progState) {
                if (! progState.runFailure && progState.runComplete) {
                    int numTiles = progState.progOutputs.length ;
                    
                    // All this code to find a nice factorization of gSize
                    
                    
                    // rSmall is the small dimention of the panel
                    // rBig is the large dimention of the panel
                    // rThin is true, if the panel is taller than it is wide
                    int rSmall  = panel.getWidth() ;
                    int rBig    = panel.getHeight() ;
                    boolean rThin = true ;
                    if (rSmall > rBig) {
                        rBig   = panel.getWidth() ;
                        rThin  = false ;
                    }
                    
                    // Generally, numTiles = gFactLowSmall*gFactLowBig
                    //                     = gFactHighSmall*gFactHighBig
                    int gFactLowSmall  = 1 ;
                    int gFactLowBig    = numTiles ;
                    int gFactHighSmall = 1 ;
                    int gFactHighBig   = numTiles ;
                    // This would be rather inefficient if numTiles were large
                    for (int i=2; i*i<=numTiles && 
                                  gFactHighSmall*rBig<gFactHighBig*rSmall ; ++i)
                        if (numTiles%i == 0) {
                            // We'return discovered another factor
                            gFactLowSmall  = gFactHighSmall ;
                            gFactLowBig    = gFactHighBig ;
                            gFactHighSmall = i ;
                            gFactHighBig   = numTiles/i ;
                        }
                    int gFactGoodSmall = gFactHighSmall ;
                    int gFactGoodBig   = gFactHighBig ;
                    
                    // Check the results
                    // if gFactHighSmall == 1,
                    //    we never found another factor
                    // if gFactHighSmall*rBig<gFactHighBig*rSmall,
                    //    we left the loop by failing the i*i<=gSize case.
                    //    In the case, the next factor would be gFactHighBig itself;
                    //    however that rectangle is too tall!
                    // if gFactHighSmall*rBig>gFactHighBig*rSmall,
                    //    then gFactLowSmall*rBig<=gFactLowBig*rSmall.
                    //    One is too small and one is too big.
                    //    Choose the one whose ratio, as shown below
                    //       gFactLowSmall/gFactLowBig
                    //       gFactHighSmall/gFactHighBig
                    //    is closest to rSmall/rBig.
                    //    We know that
                    //       gFactLowSmall/gFactLowBig < rSmall/rBig < gFactHighSmall/gFactHighBig
                    if ((gFactHighSmall != 1)
                          && (gFactHighSmall*rBig>gFactHighBig*rSmall)
                          && (((double)rSmall/rBig-(double)gFactLowSmall/gFactLowBig) < 
                             ((double)gFactHighSmall/gFactHighBig-(double)rSmall/rBig))) {
                        gFactGoodSmall = gFactLowSmall ;
                        gFactGoodBig   = gFactLowBig ;
                    }
                    int arrWidth  = gFactGoodSmall ;
                    int arrHeight = gFactGoodBig ;
                    if (!rThin) {
                        arrWidth  = gFactGoodBig ;
                        arrHeight = gFactGoodSmall ;
                    }
                    
                    // Perform some graphics computations
                    int myWidth  = panel.getWidth() ;
                    int myHeight = panel.getHeight() ;
                    int halfWidth  = arrWidth/2 ;
                    int halfHeight = arrHeight/2 ;
                    int fontSize   = Math.min((myHeight*2)/(arrHeight*3),
                                              (myWidth*3)/(arrWidth*5)) ;
                    Font myFont = new Font("SanSerif", Font.PLAIN, fontSize) ;
                    g.setFont(myFont) ;
                    FontMetrics myFM = g.getFontMetrics() ;
                    int hintHeight   = myFM.getAscent() ;
                    boolean drawHint = (myWidth/arrWidth >= 10) && 
                                       (myHeight/arrHeight >= 6) ;
                    int presX ;
                    int nextX = 0 ;
                    for (int i=0; i<arrWidth; ++i) {
                        presX = nextX ;
                        nextX = ((i+1)*myWidth+halfWidth)/arrWidth ;
                        int deltaX = nextX - presX ;
                        int presY ;
                        int nextY = 0 ;
                        for (int j=0; j<arrHeight; ++j) {
                            presY = nextY ;
                            nextY = ((j+1)*myHeight+halfHeight)/arrHeight ;
                            int deltaY = nextY - presY ;
                            int numTile = j*arrWidth+i ;
                            if (progState.progOutputs[numTile]) {
                                g.setColor( Color.RED ) ;
                                g.fillRect(presX, presY, deltaX, deltaY) ;
                                if (drawHint) {
                                    String hintString = Integer.toString(numTile) ;
                                    int offX = (deltaX-myFM.stringWidth(hintString))/2 ;
                                    int offY = (deltaY-hintHeight)/2 ;
                                    g.setColor( Color.PINK ) ;
                                    g.drawString(hintString,presX + offX,nextY - offY) ;
                                }
                            }
                        }
                    }
                } else {
                    String errMsg ;
                    if (!progState.runComplete)
                        errMsg = "Still running" ;
                    else if (progState.badIndex>0)
                        errMsg = "Bad index: " + progState.badIndex ;
                    else if (progState.runException != null)
                        errMsg = progState.runException.toString() ;
                    else
                        errMsg = "Negative size" ;
                    edu.unca.cs.LabAids.ErrorWriter.toComponent(errMsg, g, panel) ;

                }
            }
        }
    }
}
